@@ -8,7 +8,7 @@
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/storage/smgr/md.c,v 1.114 2004/12/31 22:01:13 pgsql Exp $
+ *	  $PostgreSQL: pgsql/src/backend/storage/smgr/md.c,v 1.118.2.2 2007/01/27 20:15:55 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -28,28 +28,69 @@
 #include "utils/memutils.h"
 
 
+/* interval for calling AbsorbFsyncRequests in mdsync */
+#define FSYNCS_PER_ABSORB		10
+
+/* special values for the segno arg to RememberFsyncRequest */
+#define FORGET_RELATION_FSYNC	(InvalidBlockNumber)
+#define FORGET_DATABASE_FSYNC	(InvalidBlockNumber-1)
+
+/*
+ * On Windows, we have to interpret EACCES as possibly meaning the same as
+ * ENOENT, because if a file is unlinked-but-not-yet-gone on that platform,
+ * that's what you get.  Ugh.  This code is designed so that we don't
+ * actually believe these cases are okay without further evidence (namely,
+ * a pending fsync request getting revoked ... see mdsync).
+ */
+#ifndef WIN32
+#define FILE_POSSIBLY_DELETED(err)  ((err) == ENOENT)
+#else
+#define FILE_POSSIBLY_DELETED(err)  ((err) == ENOENT || (err) == EACCES)
+#endif
+
 /*
  *	The magnetic disk storage manager keeps track of open file
  *	descriptors in its own descriptor pool.  This is done to make it
  *	easier to support relations that are larger than the operating
  *	system's file size limit (often 2GBytes).  In order to do that,
- *	we break relations up into chunks of < 2GBytes and store one chunk
- *	in each of several files that represent the relation.  See the
- *	BLCKSZ and RELSEG_SIZE configuration constants in pg_config_manual.h.
- *	All chunks except the last MUST have size exactly equal to RELSEG_SIZE
- *	blocks --- see mdnblocks() and mdtruncate().
+ *	we break relations up into "segment" files that are each shorter than
+ *	the OS file size limit.  The segment size is set by the RELSEG_SIZE
+ *	configuration constant in pg_config_manual.h.
+ *
+ *	On disk, a relation must consist of consecutively numbered segment
+ *	files in the pattern
+ *		-- Zero or more full segments of exactly RELSEG_SIZE blocks each
+ *		-- Exactly one partial segment of size 0 <= size < RELSEG_SIZE blocks
+ *		-- Optionally, any number of inactive segments of size 0 blocks.
+ *	The full and partial segments are collectively the "active" segments.
+ *	Inactive segments are those that once contained data but are currently
+ *	not needed because of an mdtruncate() operation.  The reason for leaving
+ *	them present at size zero, rather than unlinking them, is that other
+ *	backends and/or the bgwriter might be holding open file references to
+ *	such segments.  If the relation expands again after mdtruncate(), such
+ *	that a deactivated segment becomes active again, it is important that
+ *	such file references still be valid --- else data might get written
+ *	out to an unlinked old copy of a segment file that will eventually
+ *	disappear.
  *
  *	The file descriptor pointer (md_fd field) stored in the SMgrRelation
- *	cache is, therefore, just the head of a list of MdfdVec objects.
- *	But note the md_fd pointer can be NULL, indicating relation not open.
+ *	cache is, therefore, just the head of a list of MdfdVec objects, one
+ *	per segment.  But note the md_fd pointer can be NULL, indicating
+ *	relation not open.
  *
- *	Note that mdfd_chain == NULL does not necessarily mean the relation
+ *	Also note that mdfd_chain == NULL does not necessarily mean the relation
  *	doesn't have another segment after this one; we may just not have
  *	opened the next segment yet.  (We could not have "all segments are
  *	in the chain" as an invariant anyway, since another backend could
- *	extend the relation when we weren't looking.)
+ *	extend the relation when we weren't looking.)  We do not make chain
+ *	entries for inactive segments, however; as soon as we find a partial
+ *	segment, we assume that any subsequent segments are inactive.
  *
  *	All MdfdVec objects are palloc'd in the MdCxt memory context.
+ *
+ *	Defining LET_OS_MANAGE_FILESIZE disables the segmentation logic,
+ *	for use on machines that support large files.  Beware that that
+ *	code has not been tested in a long time and is probably bit-rotted.
  */
 
 typedef struct _MdfdVec
@@ -69,19 +110,22 @@ static MemoryContext MdCxt;		/* context for all md.c allocations */
  * we keep track of pending fsync operations: we need to remember all relation
  * segments that have been written since the last checkpoint, so that we can
  * fsync them down to disk before completing the next checkpoint.  This hash
- * table remembers the pending operations.	We use a hash table not because
- * we want to look up individual operations, but simply as a convenient way
- * of eliminating duplicate requests.
+ * table remembers the pending operations.	We use a hash table mostly as
+ * a convenient way of eliminating duplicate requests.
  *
  * (Regular backends do not track pending operations locally, but forward
  * them to the bgwriter.)
- *
- * XXX for WIN32, may want to expand this to track pending deletes, too.
  */
 typedef struct
 {
 	RelFileNode rnode;			/* the targeted relation */
 	BlockNumber segno;			/* which segment */
+} PendingOperationTag;
+
+typedef struct
+{
+	PendingOperationTag tag;	/* hash table key (must be first!) */
+	int			failures;		/* number of failed attempts to fsync */
 } PendingOperationEntry;
 
 static HTAB *pendingOpsTable = NULL;
@@ -114,9 +158,9 @@ mdinit(void)
 								  ALLOCSET_DEFAULT_MAXSIZE);
 
 	/*
-	 * Create pending-operations hashtable if we need it.  Currently, we
-	 * need it if we are standalone (not under a postmaster) OR if we are
-	 * a bootstrap-mode subprocess of a postmaster (that is, a startup or
+	 * Create pending-operations hashtable if we need it.  Currently, we need
+	 * it if we are standalone (not under a postmaster) OR if we are a
+	 * bootstrap-mode subprocess of a postmaster (that is, a startup or
 	 * bgwriter process).
 	 */
 	if (!IsUnderPostmaster || IsBootstrapProcessingMode())
@@ -124,14 +168,14 @@ mdinit(void)
 		HASHCTL		hash_ctl;
 
 		MemSet(&hash_ctl, 0, sizeof(hash_ctl));
-		hash_ctl.keysize = sizeof(PendingOperationEntry);
+		hash_ctl.keysize = sizeof(PendingOperationTag);
 		hash_ctl.entrysize = sizeof(PendingOperationEntry);
 		hash_ctl.hash = tag_hash;
 		hash_ctl.hcxt = MdCxt;
 		pendingOpsTable = hash_create("Pending Ops Table",
 									  100L,
 									  &hash_ctl,
-							   HASH_ELEM | HASH_FUNCTION | HASH_CONTEXT);
+								   HASH_ELEM | HASH_FUNCTION | HASH_CONTEXT);
 	}
 
 	return true;
@@ -155,21 +199,20 @@ mdcreate(SMgrRelation reln, bool isRedo)
 
 	path = relpath(reln->smgr_rnode);
 
-	fd = FileNameOpenFile(path, O_RDWR | O_CREAT | O_EXCL | PG_BINARY, 0600);
+	fd = PathNameOpenFile(path, O_RDWR | O_CREAT | O_EXCL | PG_BINARY, 0600);
 
 	if (fd < 0)
 	{
 		int			save_errno = errno;
 
 		/*
-		 * During bootstrap, there are cases where a system relation will
-		 * be accessed (by internal backend processes) before the
-		 * bootstrap script nominally creates it.  Therefore, allow the
-		 * file to exist already, even if isRedo is not set.  (See also
-		 * mdopen)
+		 * During bootstrap, there are cases where a system relation will be
+		 * accessed (by internal backend processes) before the bootstrap
+		 * script nominally creates it.  Therefore, allow the file to exist
+		 * already, even if isRedo is not set.	(See also mdopen)
 		 */
 		if (isRedo || IsBootstrapProcessingMode())
-			fd = FileNameOpenFile(path, O_RDWR | PG_BINARY, 0600);
+			fd = PathNameOpenFile(path, O_RDWR | PG_BINARY, 0600);
 		if (fd < 0)
 		{
 			pfree(path);
@@ -208,6 +251,12 @@ mdunlink(RelFileNode rnode, bool isRedo)
 	int			save_errno = 0;
 	char	   *path;
 
+	/*
+	 * We have to clean out any pending fsync requests for the doomed relation,
+	 * else the next mdsync() will fail.
+	 */
+	ForgetRelationFsyncRequests(rnode);
+
 	path = relpath(rnode);
 
 	/* Delete the first segment, or only segment if not doing segmenting */
@@ -221,12 +270,16 @@ mdunlink(RelFileNode rnode, bool isRedo)
 	}
 
 #ifndef LET_OS_MANAGE_FILESIZE
-	/* Get the additional segments, if any */
+	/* Delete the additional segments, if any */
 	if (status)
 	{
 		char	   *segpath = (char *) palloc(strlen(path) + 12);
 		BlockNumber segno;
 
+		/*
+		 * Note that because we loop until getting ENOENT, we will
+		 * correctly remove all inactive segments as well as active ones.
+		 */
 		for (segno = 1;; segno++)
 		{
 			sprintf(segpath, "%s.%u", path, segno);
@@ -256,15 +309,10 @@ mdunlink(RelFileNode rnode, bool isRedo)
  *
  *		The semantics are basically the same as mdwrite(): write at the
  *		specified position.  However, we are expecting to extend the
- *		relation (ie, blocknum is the current EOF), and so in case of
+ *		relation (ie, blocknum is >= the current EOF), and so in case of
  *		failure we clean up by truncating.
  *
  *		This routine returns true or false, with errno set as appropriate.
- *
- * Note: this routine used to call mdnblocks() to get the block position
- * to write at, but that's pretty silly since the caller needs to know where
- * the block will be written, and accordingly must have done mdnblocks()
- * already.  Might as well pass in the position and save a seek.
  */
 bool
 mdextend(SMgrRelation reln, BlockNumber blocknum, char *buffer, bool isTemp)
@@ -283,13 +331,13 @@ mdextend(SMgrRelation reln, BlockNumber blocknum, char *buffer, bool isTemp)
 #endif
 
 	/*
-	 * Note: because caller obtained blocknum by calling _mdnblocks, which
-	 * did a seek(SEEK_END), this seek is often redundant and will be
-	 * optimized away by fd.c.	It's not redundant, however, if there is a
-	 * partial page at the end of the file.  In that case we want to try
-	 * to overwrite the partial page with a full page.	It's also not
-	 * redundant if bufmgr.c had to dump another buffer of the same file
-	 * to make room for the new page's buffer.
+	 * Note: because caller obtained blocknum by calling _mdnblocks, which did
+	 * a seek(SEEK_END), this seek is often redundant and will be optimized
+	 * away by fd.c.  It's not redundant, however, if there is a partial page
+	 * at the end of the file.	In that case we want to try to overwrite the
+	 * partial page with a full page.  It's also not redundant if bufmgr.c had
+	 * to dump another buffer of the same file to make room for the new page's
+	 * buffer.
 	 */
 	if (FileSeek(v->mdfd_vfd, seekpos, SEEK_SET) != seekpos)
 		return false;
@@ -340,23 +388,22 @@ mdopen(SMgrRelation reln, bool allowNotFound)
 
 	path = relpath(reln->smgr_rnode);
 
-	fd = FileNameOpenFile(path, O_RDWR | PG_BINARY, 0600);
+	fd = PathNameOpenFile(path, O_RDWR | PG_BINARY, 0600);
 
 	if (fd < 0)
 	{
 		/*
-		 * During bootstrap, there are cases where a system relation will
-		 * be accessed (by internal backend processes) before the
-		 * bootstrap script nominally creates it.  Therefore, accept
-		 * mdopen() as a substitute for mdcreate() in bootstrap mode only.
-		 * (See mdcreate)
+		 * During bootstrap, there are cases where a system relation will be
+		 * accessed (by internal backend processes) before the bootstrap
+		 * script nominally creates it.  Therefore, accept mdopen() as a
+		 * substitute for mdcreate() in bootstrap mode only. (See mdcreate)
 		 */
 		if (IsBootstrapProcessingMode())
-			fd = FileNameOpenFile(path, O_RDWR | O_CREAT | O_EXCL | PG_BINARY, 0600);
+			fd = PathNameOpenFile(path, O_RDWR | O_CREAT | O_EXCL | PG_BINARY, 0600);
 		if (fd < 0)
 		{
 			pfree(path);
-			if (allowNotFound && errno == ENOENT)
+			if (allowNotFound && FILE_POSSIBLY_DELETED(errno))
 				return NULL;
 			ereport(ERROR,
 					(errcode_for_file_access(),
@@ -445,8 +492,8 @@ mdread(SMgrRelation reln, BlockNumber blocknum, char *buffer)
 	if ((nbytes = FileRead(v->mdfd_vfd, buffer, BLCKSZ)) != BLCKSZ)
 	{
 		/*
-		 * If we are at or past EOF, return zeroes without complaining.
-		 * Also substitute zeroes if we found a partial block at EOF.
+		 * If we are at or past EOF, return zeroes without complaining. Also
+		 * substitute zeroes if we found a partial block at EOF.
 		 *
 		 * XXX this is really ugly, bad design.  However the current
 		 * implementation of hash indexes requires it, because hash index
@@ -498,10 +545,10 @@ mdwrite(SMgrRelation reln, BlockNumber blocknum, char *buffer, bool isTemp)
 /*
  *	mdnblocks() -- Get the number of blocks stored in a relation.
  *
- *		Important side effect: all segments of the relation are opened
+ *		Important side effect: all active segments of the relation are opened
  *		and added to the mdfd_chain list.  If this routine has not been
  *		called, then only segments up to the last one actually touched
- *		are present in the chain...
+ *		are present in the chain.
  *
  *		Returns # of blocks, or InvalidBlockNumber on error.
  */
@@ -515,13 +562,16 @@ mdnblocks(SMgrRelation reln)
 	BlockNumber segno = 0;
 
 	/*
-	 * Skip through any segments that aren't the last one, to avoid
-	 * redundant seeks on them.  We have previously verified that these
-	 * segments are exactly RELSEG_SIZE long, and it's useless to recheck
-	 * that each time. (NOTE: this assumption could only be wrong if
-	 * another backend has truncated the relation.	We rely on higher code
-	 * levels to handle that scenario by closing and re-opening the md
-	 * fd.)
+	 * Skip through any segments that aren't the last one, to avoid redundant
+	 * seeks on them.  We have previously verified that these segments are
+	 * exactly RELSEG_SIZE long, and it's useless to recheck that each time.
+	 *
+	 * NOTE: this assumption could only be wrong if another backend has
+	 * truncated the relation.	We rely on higher code levels to handle that
+	 * scenario by closing and re-opening the md fd, which is handled via
+	 * relcache flush.  (Since the bgwriter doesn't participate in relcache
+	 * flush, it could have segment chain entries for inactive segments;
+	 * that's OK because the bgwriter never needs to compute relation size.)
 	 */
 	while (v->mdfd_chain != NULL)
 	{
@@ -545,11 +595,10 @@ mdnblocks(SMgrRelation reln)
 		if (v->mdfd_chain == NULL)
 		{
 			/*
-			 * Because we pass O_CREAT, we will create the next segment
-			 * (with zero length) immediately, if the last segment is of
-			 * length REL_SEGSIZE.	This is unnecessary but harmless, and
-			 * testing for the case would take more cycles than it seems
-			 * worth.
+			 * Because we pass O_CREAT, we will create the next segment (with
+			 * zero length) immediately, if the last segment is of length
+			 * RELSEG_SIZE.  While perhaps not strictly necessary, this keeps
+			 * the logic simple.
 			 */
 			v->mdfd_chain = _mdfd_openseg(reln, segno, O_CREAT);
 			if (v->mdfd_chain == NULL)
@@ -579,8 +628,8 @@ mdtruncate(SMgrRelation reln, BlockNumber nblocks, bool isTemp)
 #endif
 
 	/*
-	 * NOTE: mdnblocks makes sure we have opened all existing segments, so
-	 * that truncate/delete loop will get them all!
+	 * NOTE: mdnblocks makes sure we have opened all active segments, so
+	 * that truncation loop will get them all!
 	 */
 	curnblk = mdnblocks(reln);
 	if (curnblk == InvalidBlockNumber)
@@ -601,14 +650,17 @@ mdtruncate(SMgrRelation reln, BlockNumber nblocks, bool isTemp)
 		if (priorblocks > nblocks)
 		{
 			/*
-			 * This segment is no longer wanted at all (and has already
-			 * been unlinked from the mdfd_chain). We truncate the file
-			 * before deleting it because if other backends are holding
-			 * the file open, the unlink will fail on some platforms.
-			 * Better a zero-size file gets left around than a big file...
+			 * This segment is no longer active (and has already been
+			 * unlinked from the mdfd_chain). We truncate the file, but do
+			 * not delete it, for reasons explained in the header comments.
 			 */
-			FileTruncate(v->mdfd_vfd, 0);
-			FileUnlink(v->mdfd_vfd);
+			if (FileTruncate(v->mdfd_vfd, 0) < 0)
+				return InvalidBlockNumber;
+			if (!isTemp)
+			{
+				if (!register_dirty_segment(reln, v))
+					return InvalidBlockNumber;
+			}
 			v = v->mdfd_chain;
 			Assert(ov != reln->md_fd);	/* we never drop the 1st segment */
 			pfree(ov);
@@ -616,12 +668,12 @@ mdtruncate(SMgrRelation reln, BlockNumber nblocks, bool isTemp)
 		else if (priorblocks + ((BlockNumber) RELSEG_SIZE) > nblocks)
 		{
 			/*
-			 * This is the last segment we want to keep. Truncate the file
-			 * to the right length, and clear chain link that points to
-			 * any remaining segments (which we shall zap). NOTE: if
-			 * nblocks is exactly a multiple K of RELSEG_SIZE, we will
-			 * truncate the K+1st segment to 0 length but keep it. This is
-			 * mainly so that the right thing happens if nblocks==0.
+			 * This is the last segment we want to keep. Truncate the file to
+			 * the right length, and clear chain link that points to any
+			 * remaining segments (which we shall zap). NOTE: if nblocks is
+			 * exactly a multiple K of RELSEG_SIZE, we will truncate the K+1st
+			 * segment to 0 length but keep it. This adheres to the invariant
+			 * given in the header comments.
 			 */
 			BlockNumber lastsegblocks = nblocks - priorblocks;
 
@@ -638,8 +690,8 @@ mdtruncate(SMgrRelation reln, BlockNumber nblocks, bool isTemp)
 		else
 		{
 			/*
-			 * We still need this segment and 0 or more blocks beyond it,
-			 * so nothing to do here.
+			 * We still need this segment and 0 or more blocks beyond it, so
+			 * nothing to do here.
 			 */
 			v = v->mdfd_chain;
 		}
@@ -660,6 +712,9 @@ mdtruncate(SMgrRelation reln, BlockNumber nblocks, bool isTemp)
 
 /*
  *	mdimmedsync() -- Immediately sync a relation to stable storage.
+ *
+ * Note that only writes already issued are synced; this routine knows
+ * nothing of dirty buffers that may exist inside the buffer manager.
  */
 bool
 mdimmedsync(SMgrRelation reln)
@@ -668,7 +723,7 @@ mdimmedsync(SMgrRelation reln)
 	BlockNumber curnblk;
 
 	/*
-	 * NOTE: mdnblocks makes sure we have opened all existing segments, so
+	 * NOTE: mdnblocks makes sure we have opened all active segments, so
 	 * that fsync loop will get them all!
 	 */
 	curnblk = mdnblocks(reln);
@@ -701,82 +756,130 @@ mdimmedsync(SMgrRelation reln)
 bool
 mdsync(void)
 {
-	HASH_SEQ_STATUS hstat;
-	PendingOperationEntry *entry;
+	bool		need_retry;
 
 	if (!pendingOpsTable)
 		return false;
 
 	/*
-	 * If we are in the bgwriter, the sync had better include all fsync
-	 * requests that were queued by backends before the checkpoint REDO
-	 * point was determined.  We go that a little better by accepting all
-	 * requests queued up to the point where we start fsync'ing.
+	 * The fsync table could contain requests to fsync relations that have
+	 * been deleted (unlinked) by the time we get to them.  Rather than
+	 * just hoping an ENOENT (or EACCES on Windows) error can be ignored,
+	 * what we will do is retry the whole process after absorbing fsync
+	 * request messages again.  Since mdunlink() queues a "revoke" message
+	 * before actually unlinking, the fsync request is guaranteed to be gone
+	 * the second time if it really was this case.  DROP DATABASE likewise
+	 * has to tell us to forget fsync requests before it starts deletions.
 	 */
-	AbsorbFsyncRequests();
+	do {
+		HASH_SEQ_STATUS hstat;
+		PendingOperationEntry *entry;
+		int			absorb_counter;
+
+		need_retry = false;
 
-	hash_seq_init(&hstat, pendingOpsTable);
-	while ((entry = (PendingOperationEntry *) hash_seq_search(&hstat)) != NULL)
-	{
 		/*
-		 * If fsync is off then we don't have to bother opening the file
-		 * at all.	(We delay checking until this point so that changing
-		 * fsync on the fly behaves sensibly.)
+		 * If we are in the bgwriter, the sync had better include all fsync
+		 * requests that were queued by backends before the checkpoint REDO
+		 * point was determined. We go that a little better by accepting all
+		 * requests queued up to the point where we start fsync'ing.
 		 */
-		if (enableFsync)
+		AbsorbFsyncRequests();
+
+		absorb_counter = FSYNCS_PER_ABSORB;
+		hash_seq_init(&hstat, pendingOpsTable);
+		while ((entry = (PendingOperationEntry *) hash_seq_search(&hstat)) != NULL)
 		{
-			SMgrRelation reln;
-			MdfdVec    *seg;
-
 			/*
-			 * Find or create an smgr hash entry for this relation. This
-			 * may seem a bit unclean -- md calling smgr?  But it's really
-			 * the best solution.  It ensures that the open file reference
-			 * isn't permanently leaked if we get an error here. (You may
-			 * say "but an unreferenced SMgrRelation is still a leak!" Not
-			 * really, because the only case in which a checkpoint is done
-			 * by a process that isn't about to shut down is in the
-			 * bgwriter, and it will periodically do smgrcloseall().  This
-			 * fact justifies our not closing the reln in the success path
-			 * either, which is a good thing since in non-bgwriter cases
-			 * we couldn't safely do that.)  Furthermore, in many cases
-			 * the relation will have been dirtied through this same smgr
-			 * relation, and so we can save a file open/close cycle.
+			 * If fsync is off then we don't have to bother opening the file
+			 * at all.  (We delay checking until this point so that changing
+			 * fsync on the fly behaves sensibly.)
 			 */
-			reln = smgropen(entry->rnode);
-
-			/*
-			 * It is possible that the relation has been dropped or
-			 * truncated since the fsync request was entered.  Therefore,
-			 * we have to allow file-not-found errors.	This applies both
-			 * during _mdfd_getseg() and during FileSync, since fd.c might
-			 * have closed the file behind our back.
-			 */
-			seg = _mdfd_getseg(reln,
-							   entry->segno * ((BlockNumber) RELSEG_SIZE),
-							   true);
-			if (seg)
+			if (enableFsync)
 			{
-				if (FileSync(seg->mdfd_vfd) < 0 &&
-					errno != ENOENT)
+				SMgrRelation reln;
+				MdfdVec    *seg;
+
+				/*
+				 * If in bgwriter, we want to absorb pending requests every so
+				 * often to prevent overflow of the fsync request queue.  This
+				 * could result in deleting the current entry out from under
+				 * our hashtable scan, so the procedure is to fall out of the
+				 * scan and start over from the top of the function.
+				 */
+				if (--absorb_counter <= 0)
 				{
-					ereport(LOG,
-							(errcode_for_file_access(),
-							 errmsg("could not fsync segment %u of relation %u/%u/%u: %m",
-									entry->segno,
-									entry->rnode.spcNode,
-									entry->rnode.dbNode,
-									entry->rnode.relNode)));
-					return false;
+					need_retry = true;
+					break;
+				}
+
+				/*
+				 * Find or create an smgr hash entry for this relation. This
+				 * may seem a bit unclean -- md calling smgr?  But it's really
+				 * the best solution.  It ensures that the open file reference
+				 * isn't permanently leaked if we get an error here. (You may
+				 * say "but an unreferenced SMgrRelation is still a leak!" Not
+				 * really, because the only case in which a checkpoint is done
+				 * by a process that isn't about to shut down is in the
+				 * bgwriter, and it will periodically do smgrcloseall(). This
+				 * fact justifies our not closing the reln in the success path
+				 * either, which is a good thing since in non-bgwriter cases
+				 * we couldn't safely do that.)  Furthermore, in many cases
+				 * the relation will have been dirtied through this same smgr
+				 * relation, and so we can save a file open/close cycle.
+				 */
+				reln = smgropen(entry->tag.rnode);
+
+				/*
+				 * It is possible that the relation has been dropped or
+				 * truncated since the fsync request was entered.  Therefore,
+				 * allow ENOENT, but only if we didn't fail once already on
+				 * this file.  This applies both during _mdfd_getseg() and
+				 * during FileSync, since fd.c might have closed the file
+				 * behind our back.
+				 */
+				seg = _mdfd_getseg(reln,
+								   entry->tag.segno * ((BlockNumber) RELSEG_SIZE),
+								   true);
+				if (seg == NULL ||
+					FileSync(seg->mdfd_vfd) < 0)
+				{
+					/*
+					 * XXX is there any point in allowing more than one try?
+					 * Don't see one at the moment, but easy to change the
+					 * test here if so.
+					 */
+					if (!FILE_POSSIBLY_DELETED(errno) ||
+						++(entry->failures) > 1)
+					{
+						ereport(LOG,
+								(errcode_for_file_access(),
+								 errmsg("could not fsync segment %u of relation %u/%u/%u: %m",
+										entry->tag.segno,
+										entry->tag.rnode.spcNode,
+										entry->tag.rnode.dbNode,
+										entry->tag.rnode.relNode)));
+						return false;
+					}
+					else
+						ereport(DEBUG1,
+								(errcode_for_file_access(),
+								 errmsg("could not fsync segment %u of relation %u/%u/%u, but retrying: %m",
+										entry->tag.segno,
+										entry->tag.rnode.spcNode,
+										entry->tag.rnode.dbNode,
+										entry->tag.rnode.relNode)));
+					need_retry = true;
+					continue;	/* don't delete the hashtable entry */
 				}
 			}
-		}
 
-		/* Okay, delete this entry */
-		if (hash_search(pendingOpsTable, entry,
-						HASH_REMOVE, NULL) == NULL)
-			elog(ERROR, "pendingOpsTable corrupted");
-	}
+			/* Okay, delete this entry */
+			if (hash_search(pendingOpsTable, &entry->tag,
+							HASH_REMOVE, NULL) == NULL)
+				elog(ERROR, "pendingOpsTable corrupted");
+		}
+	} while (need_retry);
 
 	return true;
 }
@@ -798,16 +901,9 @@ register_dirty_segment(SMgrRelation reln, MdfdVec *seg)
 {
 	if (pendingOpsTable)
 	{
-		PendingOperationEntry entry;
-
-		/* ensure any pad bytes in the struct are zeroed */
-		MemSet(&entry, 0, sizeof(entry));
-		entry.rnode = reln->smgr_rnode;
-		entry.segno = seg->mdfd_segno;
-
-		if (hash_search(pendingOpsTable, &entry, HASH_ENTER, NULL) != NULL)
-			return true;
-		/* out of memory: fall through to do it locally */
+		/* push it into local pending-ops table */
+		RememberFsyncRequest(reln->smgr_rnode, seg->mdfd_segno);
+		return true;
 	}
 	else
 	{
@@ -825,25 +921,135 @@ register_dirty_segment(SMgrRelation reln, MdfdVec *seg)
  *
  * We stuff the fsync request into the local hash table for execution
  * during the bgwriter's next checkpoint.
+ *
+ * The range of possible segment numbers is way less than the range of
+ * BlockNumber, so we can reserve high values of segno for special purposes.
+ * We define two: FORGET_RELATION_FSYNC means to drop pending fsyncs for
+ * a relation, and FORGET_DATABASE_FSYNC means to drop pending fsyncs for
+ * a whole database.  (These are a tad slow because the hash table has to be
+ * searched linearly, but it doesn't seem worth rethinking the table structure
+ * for them.)
  */
 void
 RememberFsyncRequest(RelFileNode rnode, BlockNumber segno)
 {
-	PendingOperationEntry entry;
-
 	Assert(pendingOpsTable);
 
-	/* ensure any pad bytes in the struct are zeroed */
-	MemSet(&entry, 0, sizeof(entry));
-	entry.rnode = rnode;
-	entry.segno = segno;
+	if (segno == FORGET_RELATION_FSYNC)
+	{
+		/* Remove any pending requests for the entire relation */
+		HASH_SEQ_STATUS hstat;
+		PendingOperationEntry *entry;
 
-	if (hash_search(pendingOpsTable, &entry, HASH_ENTER, NULL) == NULL)
-		ereport(FATAL,
-				(errcode(ERRCODE_OUT_OF_MEMORY),
-				 errmsg("out of memory")));
+		hash_seq_init(&hstat, pendingOpsTable);
+		while ((entry = (PendingOperationEntry *) hash_seq_search(&hstat)) != NULL)
+		{
+			if (RelFileNodeEquals(entry->tag.rnode, rnode))
+			{
+				/* Okay, delete this entry */
+				if (hash_search(pendingOpsTable, &entry->tag,
+								HASH_REMOVE, NULL) == NULL)
+					elog(ERROR, "pendingOpsTable corrupted");
+			}
+		}
+	}
+	else if (segno == FORGET_DATABASE_FSYNC)
+	{
+		/* Remove any pending requests for the entire database */
+		HASH_SEQ_STATUS hstat;
+		PendingOperationEntry *entry;
+
+		hash_seq_init(&hstat, pendingOpsTable);
+		while ((entry = (PendingOperationEntry *) hash_seq_search(&hstat)) != NULL)
+		{
+			if (entry->tag.rnode.dbNode == rnode.dbNode)
+			{
+				/* Okay, delete this entry */
+				if (hash_search(pendingOpsTable, &entry->tag,
+								HASH_REMOVE, NULL) == NULL)
+					elog(ERROR, "pendingOpsTable corrupted");
+			}
+		}
+	}
+	else
+	{
+		/* Normal case: enter a request to fsync this segment */
+		PendingOperationTag key;
+		PendingOperationEntry *entry;
+		bool		found;
+
+		/* ensure any pad bytes in the hash key are zeroed */
+		MemSet(&key, 0, sizeof(key));
+		key.rnode = rnode;
+		key.segno = segno;
+
+		entry = (PendingOperationEntry *) hash_search(pendingOpsTable,
+													  &key,
+													  HASH_ENTER,
+													  &found);
+		if (!found)				/* new entry, so initialize it */
+			entry->failures = 0;
+	}
 }
 
+/*
+ * ForgetRelationFsyncRequests -- ensure any fsyncs for a rel are forgotten
+ */
+void
+ForgetRelationFsyncRequests(RelFileNode rnode)
+{
+	if (pendingOpsTable)
+	{
+		/* standalone backend or startup process: fsync state is local */
+		RememberFsyncRequest(rnode, FORGET_RELATION_FSYNC);
+	}
+	else if (IsUnderPostmaster)
+	{
+		/*
+		 * Notify the bgwriter about it.  If we fail to queue the revoke
+		 * message, we have to sleep and try again ... ugly, but hopefully
+		 * won't happen often.
+		 *
+		 * XXX should we CHECK_FOR_INTERRUPTS in this loop?  Escaping with
+		 * an error would leave the no-longer-used file still present on
+		 * disk, which would be bad, so I'm inclined to assume that the
+		 * bgwriter will always empty the queue soon.
+		 */
+		while (!ForwardFsyncRequest(rnode, FORGET_RELATION_FSYNC))
+			pg_usleep(10000L);	/* 10 msec seems a good number */
+		/*
+		 * Note we don't wait for the bgwriter to actually absorb the
+		 * revoke message; see mdsync() for the implications.
+		 */
+	}
+}
+
+/*
+ * ForgetDatabaseFsyncRequests -- ensure any fsyncs for a DB are forgotten
+ */
+void
+ForgetDatabaseFsyncRequests(Oid dbid)
+{
+	RelFileNode rnode;
+
+	rnode.dbNode = dbid;
+	rnode.spcNode = 0;
+	rnode.relNode = 0;
+
+	if (pendingOpsTable)
+	{
+		/* standalone backend or startup process: fsync state is local */
+		RememberFsyncRequest(rnode, FORGET_DATABASE_FSYNC);
+	}
+	else if (IsUnderPostmaster)
+	{
+		/* see notes in ForgetRelationFsyncRequests */
+		while (!ForwardFsyncRequest(rnode, FORGET_DATABASE_FSYNC))
+			pg_usleep(10000L);	/* 10 msec seems a good number */
+	}
+}
+
+
 /*
  *	_fdvec_alloc() -- Make a MdfdVec object.
  */
@@ -880,7 +1086,7 @@ _mdfd_openseg(SMgrRelation reln, BlockNumber segno, int oflags)
 		fullpath = path;
 
 	/* open the file */
-	fd = FileNameOpenFile(fullpath, O_RDWR | PG_BINARY | oflags, 0600);
+	fd = PathNameOpenFile(fullpath, O_RDWR | PG_BINARY | oflags, 0600);
 
 	pfree(fullpath);
 
@@ -926,29 +1132,28 @@ _mdfd_getseg(SMgrRelation reln, BlockNumber blkno, bool allowNotFound)
 		{
 			/*
 			 * We will create the next segment only if the target block is
-			 * within it.  This prevents Sorcerer's Apprentice syndrome if
-			 * a bug at higher levels causes us to be handed a
-			 * ridiculously large blkno --- otherwise we could create many
-			 * thousands of empty segment files before reaching the
-			 * "target" block.	We should never need to create more than
-			 * one new segment per call, so this restriction seems
-			 * reasonable.
+			 * within it.  This prevents Sorcerer's Apprentice syndrome if a
+			 * bug at higher levels causes us to be handed a ridiculously
+			 * large blkno --- otherwise we could create many thousands of
+			 * empty segment files before reaching the "target" block.	We
+			 * should never need to create more than one new segment per call,
+			 * so this restriction seems reasonable.
 			 *
 			 * BUT: when doing WAL recovery, disable this logic and create
-			 * segments unconditionally.  In this case it seems better
-			 * to assume the given blkno is good (it presumably came from
-			 * a CRC-checked WAL record); furthermore this lets us cope
-			 * in the case where we are replaying WAL data that has a write
-			 * into a high-numbered segment of a relation that was later
-			 * deleted.  We want to go ahead and create the segments so
-			 * we can finish out the replay.
+			 * segments unconditionally.  In this case it seems better to
+			 * assume the given blkno is good (it presumably came from a
+			 * CRC-checked WAL record); furthermore this lets us cope in the
+			 * case where we are replaying WAL data that has a write into a
+			 * high-numbered segment of a relation that was later deleted.	We
+			 * want to go ahead and create the segments so we can finish out
+			 * the replay.
 			 */
 			v->mdfd_chain = _mdfd_openseg(reln,
 										  nextsegno,
-								  (segstogo == 1 || InRecovery) ? O_CREAT : 0);
+								(segstogo == 1 || InRecovery) ? O_CREAT : 0);
 			if (v->mdfd_chain == NULL)
 			{
-				if (allowNotFound && errno == ENOENT)
+				if (allowNotFound && FILE_POSSIBLY_DELETED(errno))
 					return NULL;
 				ereport(ERROR,
 						(errcode_for_file_access(),
